;+
; NAME:
;   XFWINDOW
;
; AUTHOR:
;   Craig B. Markwardt, NASA/GSFC Code 662, Greenbelt, MD 20770
;   craigm@lheamail.gsfc.nasa.gov
;   UPDATED VERSIONs can be found on my WEB PAGE: 
;      http://cow.physics.wisc.edu/~craigm/idl/idl.html
;
; PURPOSE:
;   Intelligent replacement for built-in WINDOW procedure (Unix)
;
; MAJOR TOPICS:
;   Widgets, Graphics.
;
; CALLING SEQUENCE:
;   XFWINDOW, WINDOW_INDEX, RETAIN=retain, TITLE=title, $
;             XPOS=xpos, YPOS=ypos, XSIZE=xsize, YSIZE=ysize, [ /INIT ]
;
; DESCRIPTION:
;
;   XFWINDOW is an intelligent replacement for the normal direct
;   graphics window of IDL.  XFWINDOW uses the XFILTER library to
;   store and replay graphics commands.  This makes it easy to make
;   interactive graphics and then send the output to a postscript file
;   or printer.  
;
;   XFWINDOW is currently only usable on IDL versions 5.1 and greater,
;   under the following (Unix only) systems: SunOS, Solaris/Sparc, SGI
;   IRIX, DEC Alpha Unix and Linux.
;
;   Normally, XFWINDOW is run from the command line, and brings up a
;   graphics widget running in "NO_BLOCK" mode.  This allows you to
;   execute graphics commands and the IDL command prompt while the
;   widget is still on the screen.
;
;   All graphics commands which produce output on the window are
;   stored in memory cache.  Any command that causes the window to be
;   erased also clears the cache.  The memory cache thus represents a
;   faithful representation of the screen.  *However*, the cache does
;   not store system variables such as !P, !X, !Y, or the colormap, so
;   any changes to those variables cannot be reproduced.
;
;   The graphics window that is displayed is normal in almost every
;   sense.  Therefore all of the standard graphics commands will work
;   with it unmodified.  When you wish to delete the window, you can
;   either choose the menu "Close" selection or issue a "wdelete, win"
;   command at the prompt.
;
;   At the top of the draw window is a button bar, displayed as
;   follows:
;
;      |Configure| |Output To| o File __________ o Printer ___________
;
;   You can output the current graphics screen to a file or printer
;   (or both) queue at any time using the "Output To" button.  The
;   choice of output destinations is decided by whether the "File" or
;   "Printer" buttons are checked.  
;
;   You can choose a new file name or printer name (if not the
;   default) by filling in either the File or Printer boxes with a new
;   name.  Output is to postscript files/printers only.
;
;   The "Configure" menu allows you to perform more advanced
;   configurations.  Configure | Postscript allows you to select a
;   print layout, while Configure | Colors allows you to select a new
;   color map.
;
;   XFWINDOW is not particularly suited for situations where there are
;   many, many graphics commands comprising a plot, or if the data
;   involved in each command is very large.  In that case it may be
;   possible to exhaust IDL's memory.  In such cases it is best to
;   *not* use XFWINDOW and instead generate a script to reproduce your
;   graphics.
;
; PARAMETERS:
;
;   WINDOW_INDEX - This is an output parameter only.  XWINDOW
;                  allocates a new graphic window from the "FREE" pool
;                  (window numbers greater than 32) and returns the
;                  window number in this variable.  A value of -1
;                  indicates that the window was not successfully
;                  created.
;
;                  You *cannot* request a window number to be used.
;                  Any window number passed as input to WINDOW_INDEX
;                  is ignored and overwritten.
;
; INPUT KEYWORD PARAMETERS:
;
;   RETAIN - Specifies how the window backing store is maintained.
;            [See IDL documentation for information on backing store.]
;            Set this keyword to 0, 1 or 2 according to the following
;            table.  The default value is 1.
;
;                  0 - no backing store is provided
;                  1 - window system is requested to provide backing
;                      store (it may or may not actually provide it).
;                  2 - IDL will provide backing store.
;
;   TITLE - Specifies a the title for the new window.  By default the
;           window is given a title "XFWindow NN" where NN is the draw
;           window number.
;
;   XSIZE - Horizontal size of draw window in pixels.  The size cannot
;           be smaller than the button bar.  [ XFWINDOW will adjust
;           the size upward if needed. ]  Default is 500 pixels.
;   YSIZE - Vertical size of draw window in pixels.  Default is 400
;           pixels.
;
;   XPOS - X position of upper left hand corner of window on screen.
;   YPOS - Y position of upper left hand corner of window on screen.
;
;   NO_PRINT - set this keyword to disable the "Print" portion of the
;              window adornments.
;             NOTE:  you can't disable both "Print" and "File" parts.
;
;   NO_FILE - set this keyword to disable the "File" portion of the
;             window adornments.
;             NOTE:  you can't disable both "Print" and "File" parts.
;
;   INIT - Set this keyword to load the XFILTER shared library only.
;          No windows are displayed.
;
;   LPR - a string, the path of the Unix print command.  Default:
;         'lpr'
;
;   SCRATCH_DIR - a string, the path of a Unix scratch
;                 directory. Default: '/var/tmp'
;              
; MODIFICATION HISTORY:
;
;   Written, Aug-Sep 1999, CM
;   Merged in TVIMAGE and CMCONGRID, Sep 19, 1999
;
;-
; Copyright (C) 1999, Craig Markwardt
;
; This file is part of XFWINDOW.
;
; XFWINDOW is free software; you can redistribute it and/or modify it
; under the terms of the GNU General Public License as published by the
; Free Software Foundation; either version 2, or (at your option) any
; later version.
; 
; XFWINDOW is distributed in the hope that it will be useful, but WITHOUT
; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
; FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
; for more details.
; 
; You should have received a copy of the GNU General Public License
; along with XFWINDOW; see the file COPYING.  If not, write to the Free
; Software Foundation, 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA
;-

; Event handler: when the draw widget is deleted
pro xfwindow_draw_died, draw
  tlb = draw
  
  ;; Note the confabulations required because the widget's draw window
  ;; has already been deleted.
  NEXT_PARENT:
  newtlb = widget_info(tlb, /parent)
  if newtlb NE 0 then begin
      tlb = newtlb
      goto, NEXT_PARENT
  endif
  widget_control, tlb, $
    send_event={XFILTER_DRAW_DIED, id:tlb, top:tlb, handler:tlb, draw:draw}
  return
end

;; Make a system beep.  In versions prior to IDL 5 this was not
;; possible, but XFWINDOW is IDL v5+ only
pro xfwindow_beep
    beep
end


FUNCTION xfwindow_alert, message, XOffSet=xoff, YOffSet=yoff, parent=parent

   ; Put up a message box

IF N_PARAMS() EQ 0 THEN message = 'Please wait...'
Device, Get_Screen_Size=screenSize
IF N_ELEMENTS(xoff) EQ 0 THEN xoff = (screenSize(0)/2.0 - 100)
IF N_ELEMENTS(yoff) EQ 0 THEN yoff = (screenSize(1)/2.0 - 75)

tlb = Widget_Base(Title='Writing a File...', XOffSet=xoff, YOffSet=yoff, $
                 /floating, group_leader=parent, column=1)
label = Widget_Label(tlb, Value='')
label = Widget_Label(tlb, Value='   '+message+'   ')
label = Widget_Label(tlb, Value='')
Widget_Control, tlb, /Realize
RETURN, tlb
END ;*******************************************************************

forward_function dialog_message

;; Generate Postscript output to a file or printer
pro xfwindow_output, info

  ;; Determine whether there are commands worth printing
  call_procedure, 'xfilter_control', info.draw_window, n_commands=n
  if n_elements(n) EQ 1 then if n(0) EQ 0 then return

  ;; Get printer and file name
  filename = ''
  printer = ''
  if widget_info(info.tfile, /valid) then $
    widget_control, info.tfile,    get_value=filename
  if widget_info(info.tprinter, /valid) then $
    widget_control, info.tprinter, get_value=printer
  filename = filename(0)
  printer  = printer(0)
  info.ps_config.filename=filename
  bfile    = info.bfile
  bprinter = info.bprinter

  ;; Output was requested
  if bfile OR bprinter then begin
      ps_config = info.ps_config
      tmpdir    = info.tmpdir
      tmpfile = 0

      msg = 'Sending to ' & plural = 0
      if bfile AND ps_config.filename NE '' then begin
          msg = msg + ps_config.filename + ' '
          plural = 1
      endif
      if bprinter then begin
          if plural then msg = msg + 'and '
          if printer EQ '' then $
            msg = msg + 'Default Printer' $
          else $
            msg = msg + printer
          plural = 1
      endif
      ;; Print to a temporary file if no other filename was requested.
      if ps_config.filename EQ '' OR bfile EQ 0 then begin

          ;; This logic generates a unique name for the temporary file
          ;; based on a random seed, the current time, and a sequence
          ;; counter.  In principle this could be foiled if another
          ;; program is run at precisely the same time, but I will
          ;; assume that not *that* many people will be running this
          ;; simultaneously.
          COMMON xfwindow_tmpname, RANDOM_SEED, SEQ_COUNTER
          IF N_ELEMENTS(RANDOM_SEED) EQ 0 THEN BEGIN
              RANDOM_VAL  = LONG(SYSTIME(1))
              RANDOM_SEED = LONG(RANDOMU(RANDOM_VAL)*DOUBLE(ISHFT(1L,31)))
              SEQ_COUNTER = 0L
          ENDIF
          
          HASH = LONG(SYSTIME(1)) XOR RANDOM_SEED XOR ISHFT(SEQ_COUNTER,16)
          SEQ_COUNTER = SEQ_COUNTER + 1
          
          tmpname = tmpdir+'/xfw' + STRING(ABS(HASH), FORMAT='(Z8.8)') + '.ps'
          ps_config.filename = tmpname
          msg = msg + ' ('+tmpname+')'
          tmpfile = 1
      endif
      time0 = systime(1)
      msg = msg + '...'
      id = xfwindow_alert(msg, parent=info.tlb)

      ;; If file already exists then confirm the overwrite
      openr, unit, ps_config.filename, /get_lun, error=pserror
      if pserror EQ 0 then begin
          free_lun, unit
          qresult = dialog_message('Overwrite '+ps_config.filename+' ?', $
                                   /question, /default_no, $
                                   dialog_parent=info.tlb)
          if qresult EQ 'No' then begin
              xfwindow_beep
              return
          endif
      endif

      ;; Save the color map
      obackground = !p.background
      nc = !d.n_colors < 256L
      call_procedure, 'xfilter_control', info.draw_window, /stop
      tvlct, r0, g0, b0, /get
      nc = min([nc, n_elements(r0), n_elements(g0), n_elements(b0)])
      if nc GT 0 then $
        r = r0(0:nc-1) & g = g0(0:nc-1) & b = b0(0:nc-1)

      ;; Generate the output
      set_plot, 'ps'
      !p.background = obackground
      tvlct, r, g, b
      device, _extra=ps_config
      xfwindow_replay, info.draw_window
      device, /close
      set_plot, 'x'
      if nc GT 0 then $
        tvlct, r0, g0, b0
      !p.background = obackground
      call_procedure, 'xfilter_control', info.draw_window, /start

      ;; Print if desired
      if bprinter then begin
          lpr = info.lpr
          if printer EQ '' then $
            cmd = lpr+' '+ps_config.filename $
          else $
            cmd = lpr+' -P'+printer+' '+ps_config.filename
          spawn, cmd, dummy
      endif

      ;; Remove temporary files
      if tmpfile then $
        spawn, 'rm '+ps_config.filename, dummy

      while systime(1) - time0 LT 1.5D do begin
          wait, 0.2
      endwhile
      widget_control, id, /destroy
  endif else begin
      xfwindow_beep
  endelse
  return
end

pro xfwindow_help, print=print
  msg = ['XFWINDOW v0.8      Copyright (C) 1999 Craig B. Markwardt', $
         '', $
         'XFWINDOW is a replacement for the standard IDL direct graphics',$
         'window.  It remembers the graphics commands you used to create', $
         'your current plot, and can redirect those commands to a', $
         'postscript file or printer.', $
         '', $
         'Use XFWINDOW like a normal graphics window.  Type "XFWINDOW"', $
         'at the command line, and then issue any graphics commands you', $
         'desire.  Select either File output, Printer output, or both', $
         'by clicking the appropriate check boxes.  Fill in the appropriate',$
         'Postscript file name or printer name in the spaces provided.', $
         'Then click the "Output To" button to generate the output.', $
         '', $
         'Click the Configure|Postscript menu to change the page layout.', $
         '', $
         'XFWINDOW comes with ABSOLUTELY NO WARRANTY.  This is free', $
         'software, and you are welcome to redistribute it under the', $
         'terms of the GNU General Public License (version 2).' $
        ]
  if keyword_set(print) then begin
      print, msg, format='(A)'
  endif else begin
      ok = dialog_message(/info, msg)
  endelse
end

;; Configure the postscript page layout
pro xfwindow_ps_form, info
  filename = ''
  if widget_info(info.tfile, /valid) then $
    widget_control, info.tfile, get_value=filename
  info.ps_config.filename=filename(0)
  ;; Must do this to preserve the color tables
  call_procedure, 'xfilter_control', info.draw_window, /stop
  psnew = cmps_form(defaults=info.ps_config, button_names=['Apply'], $
                    cancel=cancel)
  call_procedure, 'xfilter_control', info.draw_window, /start
  if cancel NE 1 then begin
      info.ps_config = psnew
      if widget_info(info.tfile, /valid) then begin
          widget_control, info.tfile, update=0
          widget_control, info.tfile, set_value=psnew.filename
          widget_control, info.tfile, set_text_select=strlen(psnew.filename)
          widget_control, info.tfile, update=1
      endif
  endif
end

;; Event to File check box
pro xfwindow_bfile, event, info
  info.bfile    = (event.select EQ 1)
  widget_control, info.tfile, sensitive=info.bfile
  return
end
;; Event to Printer check box
pro xfwindow_bprinter, event, info
  info.bprinter = (event.select EQ 1)
  widget_control, info.tprinter, sensitive=info.bprinter
  return
end

;; Events to all button bar widgets
pro xfwindow_button_events, event

  widget_control, event.id, get_uvalue=uvalue
  tlb = event.top
  widget_control, tlb, get_uvalue=info, /no_copy
  case uvalue of 
      'CLOSE'      : widget_control, tlb, send_event={XFILTER_DRAW_DIED, $
                                          id:tlb, top:tlb, handler:tlb, $
                                          draw:info.draw_window}
      'OUTPUT'     : xfwindow_output, info
      'REDRAW'     : xfwindow_redraw, info.draw_window
      'HELP'       : xfwindow_help
      'POSTSCRIPT' : xfwindow_ps_form, info
      'COLORS'     : xcolors
      'BFILE'      : xfwindow_bfile,    event, info
      'BPRINTER'   : xfwindow_bprinter, event, info
      'ERASE'      : begin
          call_procedure, 'xfilter_control', info.draw_window, /erase
          erase
      end
      else         :
  end
  widget_control, tlb, set_uvalue=info, /no_copy
  
end

;; FUNCTION XFWINDOW_CONGRID2  -- two dimensional interpolation
function xfwindow_congrid2, arr, x, y
  s = size(arr)
  half = 0.5

  srx = float(s(1)/x) * findgen(x)  - half
  sry = float(s(2)/y) * findgen(y)  - half
  RETURN, INTERPOLATE(arr, srx, sry, /GRID, CUBIC=cub)
end

;; FUNCTION xfwindow_congrid3 -- three dimensional interpolation
function xfwindow_congrid3, arr, x, y, z
  s = size(arr)
  half = 0.5
 
  srx = float(s(1)/x) * findgen(x) - half
  sry = float(s(2)/y) * findgen(y) - half
  srz = float(s(3)/z) * findgen(z) - half
  RETURN, interpolate(arr, srx, sry, srz, /grid)
end

;; TVIMAGE.PRO
;; Originally Copyright David Fanning (c) 1997
;; Modified by Craig Markwardt
PRO xfwindow_tvimage, image, POSITION=position, _extra=extra

ON_ERROR, 1

   ; Check for image parameter.

np = N_PARAMS()
IF np EQ 0 THEN MESSAGE, 'You must pass an image argument.'

   ; Check image size.

s = SIZE(image)
IF s(0) LT 2 OR s(0) GT 3 THEN $
   MESSAGE, 'Argument does not appear to be an image. Returning...'

   ; 2D image.

IF s(0) EQ 2 THEN BEGIN
   imgXsize = FLOAT(s(1))
   imgYsize = FLOAT(s(2))
   true = 0
ENDIF

   ; 3D image.

IF s(0) EQ 3 THEN BEGIN
IF (s(1) NE 3L) AND (s(2) NE 3L) AND (s(3) NE 3L) THEN $
   MESSAGE, 'Argument does not appear to be a 24-bit image. Returning...'
   IF s(1) EQ 3 THEN true = 1 ; Pixel interleaved
   IF s(2) EQ 3 THEN true = 2 ; Row interleaved
   IF s(3) EQ 3 THEN true = 3 ; Band interleaved
   CASE true OF
      1: BEGIN
         imgXsize = FLOAT(s(2))
         imgYsize = FLOAT(s(3))
         END
      2: BEGIN
         imgXsize = FLOAT(s(1))
         imgYsize = FLOAT(s(3))
         END
      3: BEGIN
         imgXsize = FLOAT(s(1))
         imgYsize = FLOAT(s(2))
         END
   ENDCASE
ENDIF

   ; Check for keywords.

IF N_ELEMENTS(position) EQ 0 THEN position = [0.15, 0.15, 0.85, 0.85] $
   ELSE position = FLOAT(position)

   ; Calculate the image size and start locations.

xsize = (position(2) - position(0)) * !D.X_VSIZE
ysize = (position(3) - position(1)) * !D.Y_VSIZE
xstart = position(0) * !D.X_VSIZE
ystart = position(1) * !D.Y_VSIZE

   ; Display the image. Sizing different for PS device.

IF (!D.NAME EQ 'PS') THEN BEGIN

      ; Need a gray-scale color table if this is a true
      ; color image.

   IF true GT 0 THEN LOADCT, 0, /Silent
   TV, image, xstart, ystart, XSIZE=xsize, $
      YSIZE=ysize, _EXTRA=extra, True=true

ENDIF ELSE BEGIN

      ; If the image is 24-bit but the display is 8-bit
      ; then COLOR_QUAN processing is required.

   IF (!D.Flags AND 256) GT 0 THEN BEGIN
      thisWindow = !D.Window
      Window, XSize=10, YSize=10, /Free, /Pixmap
      WDelete, !D.Window
      WSet, thisWindow
   ENDIF
   ncolors = !D.N_Colors
   IF ncolors LE 256 AND true GT 0 THEN BEGIN
       TV, xfwindow_congrid2(COLOR_QUAN(image, true, red, green, blue, $
         Colors=!D.N_Colors), CEIL(xsize), CEIL(ysize)), $
         xstart, ystart, _EXTRA=extra
      TVLCT, red, green, blue
      RETURN
   ENDIF

   if true EQ 0 then begin
       TV, xfwindow_congrid2(image, CEIL(xsize), CEIL(ysize)), $
         xstart, ystart, _EXTRA=extra
   endif else begin
       case true of 
           1: a = [3L, ceil(xsize), ceil(ysize)]
           2: a = [ceil(xsize), 3L, ceil(ysize)]
           3: a = [ceil(xsize), ceil(ysize), 3L]
       endcase
       tv, xfwindow_congrid3(image, a(0), a(1), a(2)), $
         xstart, ystart, true=true, _extra=extra
   endelse
ENDELSE
END

;; Replacement for the IDL TV command.  Why?  Because the window size
;; may have changed since the original command and we must rescale the
;; image appropriately.
pro xfwindow_tv, img, x, y, ch, true=true, _extra=extra

  common xfwindow_replay_size, cached_xsize, cached_ysize
  if n_elements(img) EQ 0 then return
  if n_elements(x) EQ 0 then x = 0
  if n_elements(y) EQ 0 then y = 0
  if n_elements(true) EQ 0 then true = 0
  sz = size(img)
  if sz(0) LT 2 then return
  if sz(0) NE 3 AND true NE 0 then return
  ;; Deal with true-color images
  if true EQ 1 then begin
      xsize = sz(2)
      ysize = sz(3)
  endif else if true EQ 2 then begin
      xsize = sz(1)
      ysize = sz(3)
  endif else begin
      xsize = sz(1)
      ysize = sz(2)
  endelse

  ;; Generate a new position
  position  = [float(x)/cached_xsize,      float(y)/cached_ysize    ]
  dposition = [float(xsize)/cached_xsize,  float(ysize)/cached_ysize]
  position=[position, position+dposition]

  ;; Rescale the image using TVIMAGE
  xfwindow_tvimage, img, position=position, _extra=extra
  
end

;; Replacement for the IDL TV command.  Why?  Because the window size
;; may have changed since the original command and we must rescale the
;; image appropriately.
pro xfwindow_tvscl, img, x, y, ch, top=top, nan=nan, true=true, _extra=extra

  common xfwindow_replay_size, cached_xsize, cached_ysize
  if n_elements(img)  EQ 0 then return
  if n_elements(x)    EQ 0 then x = 0
  if n_elements(y)    EQ 0 then y = 0
  if n_elements(top)  EQ 0 then top = 255
  if n_elements(true) EQ 0 then true = 0
  sz = size(img)
  if sz(0) LT 2 then return
  if sz(0) NE 3 AND true NE 0 then return
  ;; Deal with true-color images
  if true EQ 1 then begin
      xsize = sz(2)
      ysize = sz(3)
  endif else if true EQ 2 then begin
      xsize = sz(1)
      ysize = sz(3)
  endif else begin
      xsize = sz(1)
      ysize = sz(2)
  endelse

  ;; Scale the image
  bimg = bytscl(img, top=top, nan=nan)

  ;; Generate a new position
  position  = [float(x)/cached_xsize,      float(y)/cached_ysize    ]
  dposition = [float(xsize)/cached_xsize,  float(ysize)/cached_ysize]
  position=[position, position+dposition]

  ;; Rescale the image using TVIMAGE
  tvimage, bimg, position=position, _extra=extra
  
end

pro xfwindow_tvlct, r, g, b, s, get=get, _extra=extra
  
  if keyword_set(get) then return
  case n_params() of 
      1: tvlct, r,          _extra=extra  ;; "V" only
      2: tvlct, r, g,       _extra=extra  ;; "V" plus start
      3: tvlct, r, g, b,    _extra=extra  ;; R, G, B only
      4: tvlct, r, g, b, s, _extra=extra  ;; R, G, B plus start
      else: return
  endcase

  return
end

pro xfwindow_call_procedure, cmd, arg, key, discard=discard

  sz = size(cmd)
  if sz(sz(0)+1) NE 7 then return
  if cmd EQ '' then return
  if n_elements(cmd) EQ 0 then return

  n_args = n_tags(arg)
  if n_args GT 5 then $
    message, 'WARNING: can''t handle more than 5 parameters for '+cmd,/info
  
  if n_args GT 0 then begin
      tn = strmid(tag_names(arg), 0, 1)
      if n_args GE 1 then if tn(0) EQ 'X' then x0 = arg.(0) else x0 = [arg.(0)]
      if n_args GE 2 then if tn(1) EQ 'X' then x1 = arg.(1) else x1 = [arg.(1)]
      if n_args GE 3 then if tn(2) EQ 'X' then x2 = arg.(2) else x2 = [arg.(2)]
      if n_args GE 4 then if tn(3) EQ 'X' then x3 = arg.(3) else x3 = [arg.(3)]
      if n_args GE 5 then if tn(4) EQ 'X' then x4 = arg.(4) else x4 = [arg.(4)]
  endif
  if keyword_set(discard) then arg = 0

  sz = size(key)   
  if sz(sz(0)+1) EQ 8 then begin ;; Keywords are present
      case n_args of 
          0: call_procedure, cmd, _extra=key
          1: call_procedure, cmd, x0, _extra=key
          2: call_procedure, cmd, x0, x1, _extra=key
          3: call_procedure, cmd, x0, x1, x2, _extra=key
          4: call_procedure, cmd, x0, x1, x2, x3, _extra=key
          5: call_procedure, cmd, x0, x1, x2, x3, x4, _extra=key
      endcase
  endif else begin               ;; No keywords are present
      case n_args of 
          0: call_procedure, cmd
          1: call_procedure, cmd, x0
          2: call_procedure, cmd, x0, x1
          3: call_procedure, cmd, x0, x1, x2
          4: call_procedure, cmd, x0, x1, x2, x3
          5: call_procedure, cmd, x0, x1, x2, x3, x4
      endcase
  endelse
  return
end

;; Replay stored commands, either to the screen or to postscript
pro xfwindow_replay, win
  common xfwindow_replay_size, cached_xsize, cached_ysize

  ;; Default error policy is to ignore errors
  catch, catcherror
  if catcherror NE 0 then begin
      catch, /cancel
      xfwindow_beep
      help, /last_message
      return
  endif

  ;; Retrieve the number of stored commands
  call_procedure, 'xfilter_control', win, n_commands=n
  if n EQ 0 then return

  ;; Retrieve the original size of the window (the size may have
  ;; changed since then)
  call_procedure, 'xfilter_control', win, xsize=xsize, ysize=ysize
  cached_xsize = xsize
  cached_ysize = ysize

  ;; For each command we need to reissue.
  for i = 0, n-1 do begin
      ;; Retrieve one command at a time.  CMD is the command (a
      ;; string), ARG is a structure of the args passed to CMD, and
      ;; KEY is an extra-style structure of keyword arguments.
      cmd = 0 & arg = 0 & key = 0
      call_procedure, 'xfilter_control', win, i, /read, cmd, arg, key
      sz = size(cmd)
      if sz(sz(0)+1) NE 7 then goto, NEXT_CMD
      if cmd EQ '' then goto, NEXT_CMD

      ;; Special cases for TV and TVSCL
      if strupcase(cmd) EQ 'TV'    then cmd = 'xfwindow_tv'
      if strupcase(cmd) EQ 'TVSCL' then cmd = 'xfwindow_tvscl'
      if strupcase(cmd) EQ 'TVLCT' then cmd = 'xfwindow_tvlct'

      xfwindow_call_procedure, cmd, arg, key, /discard
      NEXT_CMD:
  endfor
  return
end

;; Redraw the stored graphics commands to the X console
pro xfwindow_redraw, win
  owin = !d.window

  ;; If no commands are stored then do nothing
  call_procedure, 'xfilter_control', win, n_commands=n
  if n_elements(n) NE 1 then return
  if n LT 0 then return

  wset, win
  ;; Be sure to stop recording the commands momentarily so that we
  ;; don't store our replayed commands.
  call_procedure, 'xfilter_control', win, /stop
  if n EQ 0 then erase                 ;; Erase the window
  if n GT 0 then xfwindow_replay, win  ;; Replay the commands
  ;; Resume recording
  call_procedure, 'xfilter_control', win, /start

  ;; Restore the original graphics window
  if owin GE -1 then wset, owin
  return
end

;; What to do when the widget is killed: remove the XFILTER'd window.
pro xfwindow_kill, tlb
  widget_control, tlb, get_uvalue=info
  if n_elements(info) EQ 0 then return
  call_procedure, 'xfilter_control', info.draw_window, n_commands=n
  if n_elements(n) EQ 1 then if n(0) GE 0 then $
    call_procedure, 'xfilter_wdelete', info.draw_window
end


pro xfwindow_colormap

;  I may need to put these in to correct the fruity colormap changes.
;  Somehow though, these two lines don't seem to make things much
;  better.

;  tvlct, r, g, b, /get
;  tvlct, r, g, b
  return
end

pro xfwindow_event, event

  evname = tag_names(event, /structure)
  case evname of 
      'XFILTER_DRAW_DIED': begin
          ;; If the draw widget died
          widget_control, event.top, /destroy
          ;; Automatically calls XFWINDOW_KILL
      end
      'WIDGET_TRACKING': if event.enter EQ 1 then xfwindow_colormap
      ;; Window resize event
      else: begin
          widget_control, event.top, get_uvalue=info, /no_copy

          hh = info.buttonbar_height
          ww = info.buttonbar_width
          widget_control, event.top, update=0
          ;; Minimum width is tied to button bar
          widget_control, info.draw_widget, draw_xsize=event.x > ww
          widget_control, info.draw_widget, draw_ysize=event.y - hh
          widget_control, event.top, update=1
          xfwindow_redraw, info.draw_window
          
          widget_control, event.top, set_uvalue=info, /no_copy
      end
  endcase
          
  return
end

;; Locate the proper shared library to load for this architecture.
pro xfwindow_load, status=status

  status=-1
  ;; Do architecture- and release-dependent sanity checks.
  thisRelease = StrMid(!Version.Release, 0, 1)
  if !version.release LT 5.1 then begin
      message, 'ERROR: XFWINDOW required IDL 5.1 or greater.',/info
      return
  endif
  if !version.os_family NE 'unix' then begin
      message, 'ERROR: XFWINDOW only works on Unix-style platforms.',/info
      return
  endif

  ;; xf_loaded is set to 1 after the shared library has been loaded.
  ;; xf_os is the OS name, as used in the shared library file name.
  common xfwindow_loaded, xf_loaded, xf_os
  if n_elements(xf_loaded) EQ 0 then xf_loaded = 0

  if NOT xf_loaded then begin
      xf_os = !version.os
      ;; SunOS type is not specific enough.  We check for Solaris
      if xf_os EQ 'sunos' then begin
          spawn, ['uname', '-r'], osver, /NOSHELL
          if n_elements(osver) EQ 1 then if osver(0) GE 5 then $
            xf_os = 'solaris'
      endif

      ;; ******************* CHANGE LINE BELOW **********************
      libpath0 = '@XFILTER_PATH@'
      ;; ******************* CHANGE LINE ABOVE **********************
      ;; Change line above to the directory that contains xfwindow.pro

      libpath = ['.']
      if libpath0 NE '@XFILTER_PATH@' then $
        libpath = [libpath, libpath0]
      if getenv('XFILTER_PATH') NE '' then $
        libpath = [libpath, getenv('XFILTER_PATH')]

      if !path NE '' then $
        libpath = [libpath, str_sep(!path, ':')]

      libbase = 'xfilter_'+xf_os+'.so'
      for i = 0, n_elements(libpath)-1 do begin
          ;; Strip all but the root trailing slash
          while strlen(libpath(i)) GT 1 AND $
            strmid(libpath(i), strlen(libpath(i))-1, 1) EQ '/' do $
            libpath(i) = strmid(libpath(i), strlen(libpath(i))-1)

          ;; Construct file name and try to open it.
          libfile = libpath(i)+'/'+libbase
          openr, unit, libfile, /get_lun, error=liberror
          if liberror NE 0 then goto, NEXT_PATH
          free_lun, unit
          goto, GOOD_PATH
          NEXT_PATH:
      endfor

      ;; If we get here we could not find the shared library
      BAD_PATH:
      message, 'ERROR: Could not find '+libbase+'.', /info
      message, '       Does it need to be compiled for your platform?', $
        /info
      return
          
      GOOD_PATH:
      ;; Expand silly paths if needed -- some shared library stuff
      ;; won't accept anything but a full path name
      if strpos(libfile, '$') GE 0 OR strpos(libfile, '~') EQ 0 then begin
          spawn, ['ls', libfile], libfile1, /NOSHELL
          if n_elements(libfile1) NE 1 then goto, BAD_PATH
          if libfile1(0) EQ '' then goto, BAD_PATH
          libfile = libfile1
      endif

      ;; Catch shared library errors
      liberror = 0
      catch, liberror
      if liberror NE 0 then begin
          catch, /cancel
          message, 'ERROR: Could not load '+libfile+'.', /info
          help, /last_message
          return
      endif
      ;; Initialize the library
      dummy = call_external(libfile, 'xfilter_init', 0)
      set_plot, 'x'
      catch, /cancel
      xf_loaded = 1
  endif
  status = 1
end


pro xfwindow, win, retain=retain, init=init, help=help, title=title, $
              xsize=xsize, ysize=ysize, xpos=xpos, ypos=ypos, $
              print_command=lpr, scratch_dir=tmpdir, $
              no_print=no_print, no_file=no_file

  ;; Use window index as status
  win = -1

  ;; Load xfwindow shared library
  status = 0
  xfwindow_load, status=status
  if status LT 0 then return

  if keyword_set(help) then begin
      print, 'USAGE: XFWINDOW, win, [XSIZE=xsize,] [YSIZE=ysize,] ...'
      print, ''
      xfwindow_help, /print
      return
  endif
  if keyword_set(init) then return
  if n_elements(retain) EQ 0 then retain = 1
  if n_elements(xsize) EQ 0 then xsize = 500
  if n_elements(ysize) EQ 0 then ysize = 400
  if n_elements(lpr) EQ 0 then lpr = 'lpr'
  if n_elements(tmpdir) EQ 0 then tmpdir = '/var/tmp'
  if keyword_set(no_print) AND keyword_set(no_file) then begin
      message, 'ERROR: /NO_PRINT and /NO_FILE are mutually exclusive', /info
      return
  endif

  ;; Following obscurity works around expanding widget bug in IDL < 5.2
  tlb0 = widget_base(title='', column=1, xpad=0, ypad=0, $
                     tlb_frame_attr=1, /tlb_size_events, $
                     xoffset=xpos, yoffset=ypos)
  tlb  = widget_base(tlb0, column=1, xpad=0, ypad=0, /tracking)

  ;; Button bar at top of window.  Draw widget will eventually have to
  ;; be as big as the button bar.
  buttonbar = widget_base(tlb, row=1, /align_left, $
                          event_pro='xfwindow_button_events')

  config    = widget_button(buttonbar, value='Configure', /menu)
  bps_form  = widget_button(config, value='Postscript', uvalue='POSTSCRIPT')
  if getenv('LOGNAME') EQ 'craigm' OR getenv('USERNAME') EQ 'craigm' then $
    bcolors   = widget_button(config, value='Colors',   uvalue='COLORS')
  berase    = widget_button(config, value='Erase',      uvalue='ERASE', $
                            /separator)
  bredraw   = widget_button(config, value='Redraw',     uvalue='REDRAW')
  help      = widget_button(config, value='Help',       uvalue='HELP', $
                           /separator)
  bclose    = widget_button(config, value='Close',      uvalue='CLOSE')

  boutput   = widget_button(buttonbar, value='Output to', uvalue='OUTPUT')
  bfile = -1L & tfile = -1L
  bfile0 = 0
  if NOT keyword_set(no_file) then begin
      if keyword_set(no_print) then begin
          bfile     = widget_label(buttonbar, value='File', uvalue='BFILE')
      endif else begin
          dummy     = widget_base(buttonbar, column=1, /nonexclusive, $
                                  xpad=0, ypad=0)
          bfile     = widget_button(dummy, value='File', uvalue='BFILE')
      endelse
      tfile     = widget_text(buttonbar, value='', /editable, uvalue='TFILE')
      bfile0    = 1
  endif
  bprinter = -1L & tprinter = -1L
  bprint0 = 0
  if NOT keyword_set(no_print) then begin
      if keyword_set(no_file) then begin
          bprinter  = widget_label(buttonbar,value='Printer',uvalue='BPRINTER')
      endif else begin
          dummy     = widget_base(buttonbar, column=1, /nonexclusive, $
                                  xpad=0, ypad=0)
          bprinter  = widget_button(dummy, value='Printer', uvalue='BPRINTER')
      endelse
      tprinter  = widget_text(buttonbar, value='', /editable,uvalue='TPRINTER')
      if bfile0 EQ 0 then bprint0 = 1
  endif

  draw      = widget_draw(tlb, xsize=xsize(0), ysize=ysize(0), $
                          uvalue=[xsize(0), ysize(0)], retain=retain)

  widget_control, tlb0, /realize
  widget_control, draw, get_value=drawwin
  ;; This is used in case the user does, "wdelete, window"
  widget_control, draw, kill_notify='xfwindow_draw_died'

  if drawwin LT 0 then begin
      widget_control, tlb0, /destroy
      message, 'ERROR: Could not open a draw window', /info
      return
  endif

  ;; Now that the widget is realized, we get the buttonbar size and
  ;; program that back into the draw widget size.
  geom = widget_info(buttonbar, /geometry)
  buttonbar_height = round(geom.ysize)
  buttonbar_width  = round(geom.xsize)
  widget_control, tlb0, tlb_get_size=tlbsize
  widget_control, draw, draw_xsize=tlbsize(0), $
    draw_ysize=tlbsize(1)-buttonbar_height

  ;; Default plot style is landscape
  ps_config = {cmps_form_info, $
               xsize: 10., xoff: 0.5, ysize: 7.5, yoff: 10.5, $
               filename: 'xfwindow.ps', $
               inches: 1, color: 1, bits_per_pixel: 8, $
               encapsulated: 0, isolatin1: 0, landscape: 1}

  info = {draw_window: drawwin, $
          draw_widget: draw, $
          tlb: tlb0, $
          buttonbar_height: buttonbar_height, $
          buttonbar_width:  buttonbar_width, $
             bfile: bfile0,    tfile: tfile, $
          bprinter: bprint0, tprinter: tprinter, $
          ps_config: ps_config, lpr: lpr, tmpdir: tmpdir}
  widget_control, tlb0, set_uvalue=info
  ;; Reset window title 
  if n_elements(title) EQ 0 then $
    title = 'XFWindow '+strtrim(drawwin,2)
  widget_control, tlb0, tlb_set_title=title
  wset, drawwin
  call_procedure, 'xfilter_window',  drawwin
  if bfile GE 0 then    widget_control, bfile,    set_button=bfile0
  if bprinter GE 0 then widget_control, bprinter, set_button=bprint0
  if widget_info(info.tfile, /valid) then begin
      widget_control, info.tfile, update=0
      widget_control, info.tfile, set_value=ps_config.filename
      widget_control, info.tfile, set_text_select=strlen(ps_config.filename)
      widget_control, info.tfile, update=1
  endif
  if widget_info(info.tprinter, /valid) then $
    widget_control, info.tprinter, sensitive=bprint0

  xmanager, 'xfwindow', tlb0, /no_block, cleanup='xfwindow_kill'

  ;; Return the window number
  win = drawwin
end
